﻿using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;

namespace Coursework
{
    //look at direction, used to intial the cameras look at
    public enum Looking : int
    {
        FORWARD = 0,
        RIGHT,
        BACK,
        LEFT,
        UP,
        DOWN
    };

    /// <summary>
    /// Stores the properties of a camera
    /// Each camera has a position, an up direction and a point to look at
    /// 
    /// A camera can either be free (can move any where)
    /// or restricted (can only move within specified bounds)
    /// </summary>
    class Camera
    {
        //camera position in the world and where it is looking at, the cam target and its up direction
        Vector3 camPosition;
        Vector3 target;
        Vector3 lookAt;
        Vector3 up;

        //initial position
        Vector3 initialCamPosition;
        Vector3 initialCamLookAt;
        Vector3 initialUp;

        //view matrix
        Matrix viewMatrix;   

        //cam movement speeds
        float camRotationSpeedKeyboard, camRotationSpeedPad;
        float camMovementSpeed;

        //culmative rotations
        float camYaw; //y axis
        float camRoll; //z axis
        float camPitch; //x axis

        //camera settings controllers
        bool restrictedCam;

        //counter variables used to determine how much it still has to move by
        float stillToMove = 0;
        float stillToTurnClockwise = 0, stillToTurnCounterClockwise = 0;

        //the currenet direction the player is facing
        PlayerDirection playerDir;

        //trigger limits used for automated movement
        float stillToMoveLimit;
        float stillToTurnClockwiseLimit;
        float stillToTurnCounterClockwiseLimit;
        float degreesToTurn;

        /// <summary>
        /// Sets up a camera
        /// </summary>
        /// <param name="initCamPos">camera position</param>
        /// <param name="look">where the camera is looking at</param>
        /// <param name="facing">the players current direction</param>
        /// <param name="restricted">is the camera restriced or not?</param>
        public void initCam(Vector3 initCamPos, Looking look, PlayerDirection facing, bool restricted)
        {
            //error checking
            if (restrictedCam)
            {
                //if the cam is restricted the look at should be the same as the player direction
                if ((int)look != (int)facing)
                {
                    look = (Looking)facing;
                }

                //give the player a direction
                playerDir = facing;
            }

            //set up the camera
            camPosition = initCamPos;
            restrictedCam = restricted;
            setLookAt(look);

            //set the initial positions etc
            initialCamLookAt = lookAt;
            initialUp = up;
            initialCamPosition = camPosition;

            //sort out the target and viewMatrix
            target = camPosition + lookAt;
            viewMatrix = Matrix.CreateLookAt(camPosition, target, up);

            //speed of the cameras movement          
            camRotationSpeedKeyboard = 1.0f / 60.0f;
            camRotationSpeedPad = 1.0f / 30.0f;
            //general non rotation movement speed
            camMovementSpeed = 0.5f;

            //set the rotations to zero
            camYaw = 0; //y axis
            camRoll = 0; //z axis
            camPitch = 0; //x axis

            //set the limits (should really pass these in based on the model dimensions)
            stillToMoveLimit = 16.0f; //set to the length of one map element
            stillToTurnClockwiseLimit = 18.0f;
            stillToTurnCounterClockwiseLimit = stillToTurnClockwiseLimit;
            degreesToTurn = 90.0f / stillToTurnCounterClockwiseLimit; //divide the number of degrees to turn by the number of steps to do it over
        }

        /********************************************
        * General camera functions
        * *****************************************/

        /// <summary>
        /// used to intialize the cameras look at and up direction
        /// 
        /// direction limited to the axis
        /// </summary>
        /// <param name="look">direction for the camera to look at</param>
        public void setLookAt(Looking look)
        {
            //check the value passes
            switch (look)
            {
                //in each case, assign the look at the corresponding value
                //and ensure the camera's up direction is the correct way up
                case Looking.BACK:
                    lookAt = Vector3.Backward;
                    up = Vector3.Up;
                break;

                case Looking.DOWN:
                    lookAt = Vector3.Down;
                    up = Vector3.Forward;
                break;

                case Looking.FORWARD:
                    lookAt = Vector3.Forward;
                    up = Vector3.Up;
                break;

                case Looking.LEFT:
                    lookAt = Vector3.Left;
                    up = Vector3.Up;
                break;

                case Looking.RIGHT:
                    lookAt = Vector3.Right;
                    up = Vector3.Up;
                break;

                case Looking.UP:
                    lookAt = Vector3.Up;
                    up = Vector3.Forward;
                break;

                default:
                    lookAt = Vector3.Forward;
                    up = Vector3.Up;
                break;
            }
        }

        //update the camera
        public void update()
        {

            //check to see if the cam's movement is restricted or not
            if (restrictedCam)
            {
                //check to see if there are any movements still to perform
                if (stillToMove > 0)
                {
                    walkForward();
                    stillToMove--;
                }
                
                if (stillToTurnClockwise > 0)
                {
                    turnClockwise();
                    stillToTurnClockwise--;
                }

                if (stillToTurnCounterClockwise > 0)
                {
                    turnCounterClockwise();
                    stillToTurnCounterClockwise--;
                }
            }

            //update the view matrix
            updateViewMatrix();
        }
       
        //applies a rotation to camera
        public void applyRotation()
        {    
            //calculate the yaw rotation from the yaw value
            Matrix yawMatrix = Matrix.CreateFromAxisAngle(Vector3.UnitY, camYaw);
            //since we are rotating, the camera's look at must be recalculated
            lookAt = Vector3.Transform(lookAt, yawMatrix);
            //depending on our current orientation, the up vector needs to be recalculated
            up = Vector3.Transform(up, yawMatrix);
            //reset the yaw rotation
            camYaw = 0;

            //code as above but for roll and pitch
            Matrix rollMatrix = Matrix.CreateFromAxisAngle(Vector3.UnitZ, camRoll);
            lookAt = Vector3.Transform(lookAt, rollMatrix);
            up = Vector3.Transform(up, rollMatrix);
            camRoll = 0;

            Matrix pitchMatrix = Matrix.CreateFromAxisAngle(Vector3.UnitX, camPitch);
            lookAt = Vector3.Transform(lookAt, pitchMatrix);
            up = Vector3.Transform(up, pitchMatrix);
            camPitch = 0;

            //update the view matrix
            updateViewMatrix();
        }

        //updates the view matrix
        //call everytime the camera changes
        public void updateViewMatrix()
        {
            //create the camera target
            target = camPosition + lookAt;
            //set the view matrix
            viewMatrix = Matrix.CreateLookAt(camPosition, target, up);
        }

        //move the cameras vertical and horizontal look at position
        public void lookUp() { lookAt.Y += camRotationSpeedKeyboard; applyRotation(); }
        public void lookDown() { lookAt.Y -= camRotationSpeedKeyboard; applyRotation(); }
        public void lookLeft() { camYaw += camRotationSpeedKeyboard; applyRotation(); }
        public void lookRight() { camYaw -= camRotationSpeedKeyboard; applyRotation(); }

        //change the camera's look at horizontal and vertical position
        //used by the gamepads thumbsticks which give a value, rather than an on/off value
        public void lookHorizontal(float rotation) { camYaw -= rotation * camRotationSpeedPad; applyRotation(); }
        public void lookVertical(float rotation) {lookAt.Y -= rotation * camRotationSpeedPad; applyRotation(); }

        //returns the camera view matrix
        public Matrix getViewMatrix() { return viewMatrix; }
        //returns the look at of the camera
        public Vector3 getTarget() { return lookAt; }
        //get set the type of camera (restricted or free)
        public void setRestrictedState(bool state) { restrictedCam = state; }
        public bool getRestrictedState() { return restrictedCam; }
        //get/set the camera position
        public void setCamPosition(Vector3 pos) { camPosition = pos; }
        public Vector3 getCamPosition() { return camPosition; }


        /********************************************
         * Free camera functions
         * *****************************************/

        //move the camera forward or back (according to its up direction)
        public void moveForward()
        {   
            camPosition += lookAt * camMovementSpeed;
            updateViewMatrix();
        }
        public void moveBack()
        {
            camPosition -= lookAt * camMovementSpeed;
            updateViewMatrix();
        }
        public void moveForward(float distance)
        {
            //seperate function required since the pads thumbsticks are used
            //thumbsticks give value rather than a on/off like a button
            camPosition += distance * lookAt * camMovementSpeed;
            updateViewMatrix();
        }
        
        //move the camera left or right (according to its up direction)
        //cross product used to work out where to strafe to
        public void strafeLeft()
        {
            camPosition -= Vector3.Cross(lookAt, up) * camMovementSpeed;
            updateViewMatrix();
        }
        public void strafeRight()
        {
            camPosition += Vector3.Cross(lookAt, up) * camMovementSpeed;
            updateViewMatrix();
        }
        public void strafe(float distance)
        {
            //seperate function required since the pads thumbsticks are used
            //thumbsticks give value rather than a on/off like a button
            camPosition += distance * Vector3.Cross(lookAt, up) * camMovementSpeed;
            updateViewMatrix();
        }

        //move the camera up or down (according to its up direction)
        public void moveUp()
        {
            camPosition += up * camMovementSpeed;
            updateViewMatrix();
        }
        public void moveDown()
        {
            camPosition -= up * camMovementSpeed;
            updateViewMatrix();
        }

        //resets the camera back to its initial position
        public void resetCam()
        {
            //only reset free cam
            if (!restrictedCam)
            {
                //back to initial values
                camPosition = initialCamPosition;
                lookAt = initialCamLookAt;
                up = initialUp;
                target = camPosition + lookAt;
                //always update the view matrix
                viewMatrix = Matrix.CreateLookAt(camPosition, target, up);
            }
        }

        /********************************************
         * Restricted camera functions
         * *****************************************/

        //moves the camera forward
        public void walkForward()
        {
            //since this is for the restricted camera we can update the camera position based on the players direction
            switch (playerDir)
            {
                //check the direction and apply the movement in the correct direction
                case PlayerDirection.NORTH:
                    camPosition.Z -= camMovementSpeed;
                break;

                case PlayerDirection.EAST:
                    camPosition.X += camMovementSpeed;
                break;

                case PlayerDirection.SOUTH:
                    camPosition.Z += camMovementSpeed;
                break;

                case PlayerDirection.WEST:
                    camPosition.X -= camMovementSpeed;
                break;
            }
                    
        }

        //rotates the camera counter clockwise
        public void turnCounterClockwise()
        {
            //radians apparently give a smoother turn than degree...cant remember where i read that
            Matrix rot = Matrix.CreateFromAxisAngle(up, MathHelper.ToRadians(degreesToTurn));
            //we are changing where the camera is looking at so update it
            lookAt = Vector3.Transform(lookAt, rot);
        }

        //rotates the camera clockwise
        public void turnClockwise()
        {
            //apply rotation
            Matrix rot = Matrix.CreateFromAxisAngle(up, MathHelper.ToRadians(-degreesToTurn));
            //update look at
            lookAt = Vector3.Transform(lookAt, rot);
        }

        //returns how long the camera still has to move forward for
        public float getStillToMove() { return stillToMove; }
        //tells the camera to move forward in the direction passed
        public void setStillToMove(PlayerDirection d)
        {
            //used to ensure the camera class and map class are both moving in the same direction
            playerDir = d;

            //the value still to move is the models length over the camera's movement speed
            stillToMove = stillToMoveLimit / camMovementSpeed;

        }

        //returns how long the camera still has to turn for
        public float getStillToTurnClockwise() { return stillToTurnClockwise; }
        public float getStillToTurnCounterClockwise() { return stillToTurnCounterClockwise; }

        //set a duration for the camera to rotate for
        public void setStillToTurnClockwise() { stillToTurnClockwise = stillToTurnClockwiseLimit; }
        public void setStillToTurnCounterClockwise() { stillToTurnCounterClockwise = stillToTurnCounterClockwiseLimit; }

    }
}

