//code adapted from http://rbwhitaker.wikidot.com and http://blogs.microsoft.co.il/blogs/tamir/archive/2008/06/17/hlsl-pixel-shader-effects-tutorial.aspx

//texture for sprite batch to draw with
texture ImageTexture;

//sampler for texture
sampler TextureSampler = sampler_state
{
	Texture = <ImageTexture>;
};

//pixel shader
//gets the colour of the texture
//black and white
float4 PixelShaderFunctionBlackAndWhite(float2 TextureCoordinate:TEXCOORD0) : COLOR0
{   
    //black and white(intensity)
    //change the rgb value by a weighting
	float4 colour = tex2D(TextureSampler, TextureCoordinate);
	colour.rgb = dot(colour.rgb, float3(0.3, 0.59, 0.11)); 
	colour.a = 1.0f;
	return colour;
}

//sepia
float4 PixelShaderFunctionSepia(float2 TextureCoordinate:TEXCOORD0) : COLOR0
{
    //sepia
    float4 colour = tex2D(TextureSampler, TextureCoordinate);
    
    float4 outputColour = (0.0f, 0.0f, 0.0f, 1.0f);
    //use a weighted combination of each colour
    outputColour.r = (colour.r * 0.393) + (colour.g * 0.769) + (colour.b * 0.189);
    outputColour.g = (colour.r * 0.349) + (colour.g * 0.686) + (colour.b * 0.168);    
    outputColour.b = (colour.r * 0.272) + (colour.g * 0.534) + (colour.b * 0.131);
	
	return outputColour;
}

//invert
float4 PixelShaderFunctionInvert(float2 TextureCoordinate:TEXCOORD0) : COLOR0
{   
   //invert
   float threshold = 0.7f;
   //take the input colour value away from the threshold value to invert
   float4 colour = threshold - tex2D(TextureSampler, TextureCoordinate);
   //ensure the value is within range
   if (colour.r < 0.0f)
		colour = 0.0f;
   if (colour.g < 0.0f)
		colour = 0.0f;
   if (colour.b < 0.0f)
		colour = 0.0f;
   //alpha needs to more than 0 to show the colour
   colour.a = 1.0f;
   
   return colour;    
}

//night vision
float4 PixelShaderFunctionNightVision(float2 TextureCoordinate:TEXCOORD0) : COLOR0
{    
	//night vision
	//get the input colour from the texture passed
	float4 outputColour = tex2D(TextureSampler, TextureCoordinate); 
	//set the coordinate colour to a near colour and apply a weighting to it
	//emphasis on the the green value
	outputColour.r -= tex2D( TextureSampler , TextureCoordinate + (10/100)).r;
	outputColour.g += tex2D( TextureSampler , TextureCoordinate + (10/200)).g;
	outputColour.b -= tex2D( TextureSampler , TextureCoordinate + (10/100)).b;

	//return the calculated colour
	return outputColour; 
}

//blur distance is used to select a near by pixel
//texture is between 0 and 1 so use a small value to prevent too much blurring
float blurDistance = 0.007f;

//blur
//blurs horizontally and vertically
float4 PixelShaderFunctionBlur(float2 TextureCoordinate:TEXCOORD0) : COLOR0
{ 
	//colour to output	
	float4 colour;	
	
	//use the blur distance to select a texture coordinate above, below, to the left and to the right
	//of the current coordinate
	//add all the values together
	colour = tex2D(TextureSampler, float2(TextureCoordinate.x+blurDistance, TextureCoordinate.y+blurDistance));
    colour += tex2D(TextureSampler, float2(TextureCoordinate.x-blurDistance, TextureCoordinate.y-blurDistance));
    colour += tex2D(TextureSampler, float2(TextureCoordinate.x+blurDistance, TextureCoordinate.y-blurDistance));
    colour += tex2D(TextureSampler, float2(TextureCoordinate.x-blurDistance, TextureCoordinate.y+blurDistance));
    
    //divide by four to get the average colour
    colour = colour/4;
    
    //return the calculated colour
    return colour;
}

technique BlackAndWhite
{
    pass Pass1
    {
            PixelShader = compile ps_2_0 PixelShaderFunctionBlackAndWhite();
    }
    
}

technique Sepia
{
    pass Pass1
    {
            PixelShader = compile ps_2_0 PixelShaderFunctionSepia();
    }
}


technique Invert
{
    pass Pass1
    {
            PixelShader = compile ps_2_0 PixelShaderFunctionInvert();
    }
    
}


technique NightVision
{
    pass Pass1
    {
            PixelShader = compile ps_2_0 PixelShaderFunctionNightVision();
    }
}

technique Blur
{
    pass Pass1
    {
            PixelShader = compile ps_2_0 PixelShaderFunctionBlur();
    }
}
