//code adapted from http://rbwhitaker.wikidot.com 

//world, view, projection matrix
float4x4 World;
float4x4 View;
float4x4 Projection;
float4x4 WorldInverseTranspose; //used for normal calculations

//general scene light 
float4 AmbientColor = float4(0.4, 0.4, 0.4, 1);
float AmbientIntensity = 0.2;
 
//light from the model
float3 DiffuseLightDirection = float3(0, 10, 0);
float4 DiffuseColor = float4(1, 1, 1, 1);
float DiffuseIntensity = 1.0;
float DiffuseLowerLimit = 0.25;

//diffuse texture input
texture ModelTexture;
//sets the sample state for the diffuse texture
sampler2D textureSampler = sampler_state 
{
    Texture = (ModelTexture);
    MinFilter = Linear;
    MagFilter = Linear;
    AddressU = Clamp;
    AddressV = Clamp;
};
 
//normal texture input
texture NormalMap;
//sets the sample state for the normal texture
//should be the same as the diffuse sampler
sampler2D bumpSampler = sampler_state 
{
    Texture = (NormalMap);
    MinFilter = Linear;
    MagFilter = Linear;
    AddressU = Clamp;
    AddressV = Clamp;
};
 
//vertex shader structure to hold the texture/vertex input information
struct VertexShaderInput
{
    float4 Position : POSITION0;
    float3 Normal : NORMAL0;
    float3 Tangent : TANGENT0;
    float3 Binormal : BINORMAL0;
    float2 TextureCoordinate : TEXCOORD0;
};

//vertex shader structure to hold the texture/vertex output information
struct VertexShaderOutput
{
    float4 Position : POSITION0;
    float2 TextureCoordinate : TEXCOORD0;
    float3 Normal : TEXCOORD1;
    float3 Tangent : TEXCOORD3; //tangent and binormal are swapped so the light comes from the right direction
    float3 Binormal : TEXCOORD2;
};

//main vertex shader function which takes in a VertexShaderInput structure and outputs a VertexShaderOutput
VertexShaderOutput VertexShaderFunction(VertexShaderInput input)
{
    VertexShaderOutput output;
	
	//calculate the worldviewprojection matrix
    float4 worldPosition = mul(input.Position, World);
    float4 viewPosition = mul(worldPosition, View);
    output.Position = mul(viewPosition, Projection);
 
	//normalize the normal, tangent and binormal. tangents and binormals required for bump mapping
    output.Normal = normalize(mul(input.Normal, WorldInverseTranspose));
    output.Tangent = normalize(mul(input.Tangent, WorldInverseTranspose));
    output.Binormal = normalize(mul(input.Binormal, WorldInverseTranspose));
 
	//grab the texture
    output.TextureCoordinate = input.TextureCoordinate;
    
    //output the structure
    return output;
}

float Bumpiness = 0.15; //determines how much bump is applied
//main pixel shader function which takes in a VertexShaderOutput structure and returns a float4 colour
float4 PixelShaderFunction(VertexShaderOutput input) : COLOR0
{
    //use the texture to work out the bumpiness
    float3 bump = Bumpiness * (tex2D(bumpSampler, input.TextureCoordinate) - (0.5, 0.5, 0.5)); //values subtracted here to ensure the value is in range
    float3 bumpNormal = input.Normal + (bump.x * input.Tangent + bump.y * input.Binormal);
    //normalize the bump value
    bumpNormal = normalize(bumpNormal);
 
    //work out the diffuse intensity using the bump map
    float diffuseIntensity = dot(normalize(DiffuseLightDirection), bumpNormal);
    //if it is lower than the minimun value, set it to the minumum
    if(diffuseIntensity < DiffuseLowerLimit)
        diffuseIntensity = DiffuseLowerLimit; 
    
    //grab the texture colour
    float4 textureColor = tex2D(textureSampler, input.TextureCoordinate);
    //ensure the alpha value is one so we can see the texture
    textureColor.a = 1;
 
    //combine diffuse and ambient colours (specular not needed since you dont get shiny bricks)
    return saturate(textureColor * (diffuseIntensity) + AmbientColor * AmbientIntensity);
}
 
technique BumpMap
{
    pass Pass1
    {
        VertexShader = compile vs_1_1 VertexShaderFunction();
        PixelShader = compile ps_2_0 PixelShaderFunction();
    }
}
